\ combined words (binding of semantics at parse time)

\ This file is in the public domain. NO WARRANTY.

\ This file is based on ideas posted in comp.lang.forth in 1998 by
\ Bernd Paysan, Jonah Thomas, Philip Preston and me.

\ Message-IDs: <70b3vc$t4n$2@newsreader2.core.theplanet.net>
\ <351B70D4.F31@remove.muenchen.this.org.junk> and their threads.

true [IF]
\ This version is standard, but not particularly efficient (this
\ affects the speed of ', POSTPONE etc. when they parse their
\ arguments).  If you want it faster, use a system-specific version
\ (e.g., the one below) or ask me for a faster version.

variable i/c-list \ contains the xt of the latest combined word
0 i/c-list !

: current' ( "name" -- xt )
    \ gets an xt for "name" in the current wordlist
    [ s" SEARCH-ORDER" environment? ] [IF]
	[IF] \ protect against CURRENT not being on top of the search order
	    bl word count get-current search-wordlist drop
	[ELSE] \ no search order wordset, current is the search order
	    '
	[THEN]
    [ELSE]
	\ don't know about search order wordset, let's assume it does
	\ not exist
	'
    [THEN] ;

: interpret/compile: ( xt-i xt-c "name" -- )
    \ creates a combined word with compilation semantics xt-c and
    \ interpretation semantics xt-i
    >in @ create immediate >in ! \ create "name" without consuming "name"
    swap , ,            \ now remember the xts
    i/c-list @ ,        \ link to previous combined word
    current' i/c-list ! \ put xt of "name" in i/c-list
DOES> ( interpretation: ... -- ... )
      ( compilation: ... -- ... )
    \ this looks like defining a STATE-smart word, but we change ',
    \ POSTPONE etc. to treat such words correctly as combined words.
    state @ IF
	cell+
    THEN
    @ execute ;

: is-i/c? ( xt -- flag )
    i/c-list begin ( xt addr -- )
	@ dup
    while
	2dup = if
	    2drop true EXIT
	then
	>body cell+ cell+
    repeat
    2drop false ;

[ELSE]
\ This variant is shorter and faster, but system dependent. It should
\ run on most systems where XTs are represented by CFAs.
\ It tested ok on Gforth and PFE, but failed on iForth (as expected).

: interpret/compile: ( xt-i xt-c "name" -- )
    Create immediate swap , ,
DOES> ( interpretation: ... -- ... )
      ( compilation: ... -- ... )
    state @ IF  cell+  THEN  @ execute ;

' dup dup interpret/compile: i/c-prototype

: is-i/c? ( xt -- flag )
  [ 0 >body 1 chars / ] literal 
  ['] i/c-prototype over compare 0= ;
[THEN]

: i/c>int ( xt1 -- xt2 )
  >body @ ;
: i/c>comp ( xt1 -- xt2 )
  >body cell+ @ ;

: literal ( compilation: n -- ; run-time: -- n )
    \ to avoid poisoning the new ['] and POSTPONE with a STATE-smart LITERAL
    \ we define a STATE-dumb LITERAL
    state @ if \ STATE-smartness to achieve state-dumbness :-)
	postpone literal
    else
	] postpone literal postpone [
    then ; immediate

: ' ( "name" -- xt )
  ' dup is-i/c? IF  i/c>int  THEN ;

\ as a special gift to Mitch Bradley this version of ['] is combined
' ' ( "name" -- xt )
:noname ( compilation: "name" -- )
        ( run-time: -- xt )
    ' postpone literal ;
interpret/compile: [']    

: comp' ( "name" -- w xt )
  bl word find dup 0= -13 and throw ( xt n)
  over is-i/c? if
    swap i/c>comp swap
  then
  0< if \ not immediate
    ['] COMPILE,
  else
    ['] EXECUTE
  then ;

: postpone, ( w xt -- )
   \ this can be optimized
   swap POSTPONE literal compile, ;

: postpone ( compilation: "name" -- )
           ( run-time: ... -- ... )
  comp' postpone, ; immediate

: [compile] ( compilation: "name" -- )
            ( run-time: ... -- ... )
  comp' drop compile, ; immediate
